/*
 * QA Remote Shell - client side
 *
 * Run a command on the server with lots of wizz-bang options
 *
 * Copyright © 2005-2008  Red Hat, Inc. All rights reserved.
 *
 * This copyrighted material is made available to anyone wishing to use,
 * modify, copy, or redistribute it subject to the terms and conditions of the
 * GNU General Public License v.2.  This program is distributed in the hope
 * that it will be useful, but WITHOUT ANY WARRANTY expressed or implied,
 * including the implied warranties of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA. Any Red Hat
 * trademarks that are incorporated in the source code or documentation are not
 * subject to the GNU General Public License and may only be used or replicated
 * with the express permission of Red Hat, Inc.
 *
 * Red Hat Author(s): Nathan Straz <nstraz@redhat.com>
 *                    Dean Jansa <djansa@redhat.com>
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <netinet/in.h>
#include <netdb.h>
#include <syslog.h>
#include <pwd.h>
#include <time.h>

#include "sockutil.h"
#include "qarsh_packet.h"
#include "btime.h"
#include "hbeat.h"

#define QARSH_MINPORT 5010
#define QARSH_BUFSIZE 4096


/* Globals */
int qarsh_fd = -1;		/* The control connection to qarshd */
hbeat_t qarsh_hb;       /* Heartbeat handle */
int signal_to_send = 0;
int sigs_to_propogate[] = { SIGINT, SIGTERM, SIGHUP, SIGUSR1, SIGUSR2 };
sigset_t pselect_sigmask;

void
usage()
{
	fprintf(stderr, "qarsh [options] [user[.group]@]hostname cmdline ...\n"
		     "-l user         Run cmdline using this user name.\n"
		     "-g group        Run cmdline using this group name.\n"
		     "-p port         Use this port to contact qarshd.\n"
		     "-t timeout      Number of seconds a remote host can be\n"
		     "                silent before we give up and exit\n"
		     "                A value of 0 disables heartbeating.\n"
		     "                Default is value is 120.\n"
		     "                Env var QARSH_TIMEOUT can also be set.\n"
	);

	return;
}

char *
copyargs(char **argv)
{
	int cc;
	char **ap, *p;
	char *args;

	cc = 0;
	for (ap = argv; *ap; ++ap)
		cc += strlen(*ap) + 1;
	if (cc == 0) return NULL;
	args = malloc(cc);
	if (!args) {
		perror("qarsh: malloc failed in copyargs");
		exit(1);
	}
	for (p = args, ap = argv; *ap; ++ap) {
		for (p = strcpy(p, *ap); *p; ++p);
		if (ap[1])
			*p++ = ' ';
	}
	return args;
}

void
sig_handler(int sig)
{
	signal_to_send = sig;
}

void 
sig_alrm_handler(int sig)
{
	fprintf(stderr, "Could not connect to remote host\n");
	exit(1);
}

void
setup_signals(void)
{
	struct sigaction sa;
	sigset_t sigmask;
	int i, n;

	n = sizeof sigs_to_propogate / sizeof *sigs_to_propogate;
	sigemptyset(&sigmask);
	for (i = 0; i < n; i++) {
		sigaddset(&sigmask, sigs_to_propogate[i]);
	}
	sigprocmask(SIG_BLOCK, &sigmask, &pselect_sigmask);
	sa.sa_handler = sig_handler;
	sa.sa_mask = sigmask;
	sa.sa_flags = SA_RESTART;
	for (i = 0; i < n; i++) {
		sigaction(sigs_to_propogate[i], &sa, NULL);
	}
}

void
reset_signals(void)
{
	struct sigaction sa;
	sigset_t sigmask;
	int i, n;

	n = sizeof sigs_to_propogate / sizeof *sigs_to_propogate;
	sigemptyset(&sigmask);
	for (i = 0; i < n; i++) {
		sigaddset(&sigmask, sigs_to_propogate[i]);
	}
	sigprocmask(SIG_UNBLOCK, &sigmask, &pselect_sigmask);
	sa.sa_handler = SIG_DFL;
	sa.sa_mask = sigmask;
	for (i = 0; i < n; i++) {
		signal(sigs_to_propogate[i], SIG_DFL);
	}
}

void
set_remote_user(char *user, char *group)
{
	struct qa_packet *qp;

	qp = make_qp_setuser(user, group);
	qp->qp_seq = 1;
	send_packet(qarsh_fd, qp);
	qpfree(qp);
	qp = recv_packet(qarsh_fd);
	if (!qp) {
		fprintf(stderr, "Failed to receive response to set user\n");
		close(qarsh_fd);
		exit(125);
	} else if (qp->qp_type == QP_RETURNCODE 
			&& qp->qp_returncode.qp_rc == -1) {
		
		fprintf(stderr, "Remote side failed, %s\n",
				qp->qp_returncode.qp_strerror);
		close(qarsh_fd);
		exit(125);
	}
}

int
run_remote_cmd(char *cmdline)
{
	struct qa_packet *qp;
	int rc;
	int p_in, p_out, p_err;  /* Port numbers */
	int l_in, l_out, l_err;  /* listening sockets */
	int c_in, c_out, c_err;  /* client sockets */
	char *b_in, *b_out, *b_err; /* Buffers */
	int z_in, z_out, z_err; /* size in buffer */
	fd_set readfds, testfds, testwds;
	int nset;
	struct sockaddr_in caddr;
	socklen_t clen;
	struct timespec timeout;
	short cmd_finished;

	l_in = bind_any(QARSH_MINPORT);
	p_in = getsockport(l_in);
	l_out = bind_any(QARSH_MINPORT);
	p_out = getsockport(l_out);
	l_err = bind_any(QARSH_MINPORT);
	p_err = getsockport(l_err);
	
	qp = make_qp_runcmd(cmdline, p_in, p_out, p_err);
	qp->qp_seq = 1;
	send_packet(qarsh_fd, qp);
	qpfree(qp);

	/* Get the stdin, stdout, and stderr connections up before we do work */
	FD_ZERO(&readfds);
	FD_SET(l_in, &readfds);
	FD_SET(l_out, &readfds);
	FD_SET(l_err, &readfds);
	c_in = c_out = c_err = 0;

	do {
		testfds = readfds;
		nset = select(FD_SETSIZE, &testfds, NULL, NULL, NULL);

		if (FD_ISSET(l_in, &testfds)) {
			clen = sizeof caddr;
			c_in = accept(l_in, (struct sockaddr *)&caddr, &clen);
			if (c_in == -1) {
				fprintf(stderr, 
					"accept of l_in failed, %d: %s\n",
				 		errno, strerror(errno));
				continue;
			}
		}
		if (FD_ISSET(l_out, &testfds)) {
			clen = sizeof caddr;
			c_out = accept(l_out, (struct sockaddr *)&caddr, &clen);
			if (c_out == -1) {
				fprintf(stderr, 
					"accept of l_out failed, %d: %s\n",
				 		errno, strerror(errno));
				continue;
			}
		}
		if (FD_ISSET(l_err, &testfds)) {
			clen = sizeof caddr;
			c_err = accept(l_err, (struct sockaddr *)&caddr, &clen);
			if (c_err == -1) {
				fprintf(stderr, 
					"accept of l_err failed, %d: %s\n",
				 		errno, strerror(errno));
				continue;
			}
		}
	} while (c_in == 0 || c_out == 0 || c_err == 0);
	close(l_in);
	close(l_out);
	close(l_err);
	l_in = l_out = l_err = -1;

	/* Now we can start doing some real work */
	
	FD_ZERO(&readfds);
	FD_SET(qarsh_fd, &readfds);
	FD_SET(c_out, &readfds);
	FD_SET(c_err, &readfds);
	/* Ignore stdin if it's a tty, we don't want to deal with
	 * tty i/o if we're a background process */
	if (!isatty(fileno(stdin))) {
		FD_SET(fileno(stdin), &readfds);
	}
	/* Setup signal handling stuff so we can propogate signals */
	setup_signals();

	if (fcntl(fileno(stdin), F_SETFL, O_NONBLOCK) != 0) {
		fprintf(stderr, 
			"fcntl stdin O_NONBLOCK failed, %d: %s\n", 
			errno, strerror(errno));
	}
	/* allocate one buffer for each stream */
	b_in = malloc(QARSH_BUFSIZE); memset(b_in, 0, QARSH_BUFSIZE); z_in = 0;
	b_out = malloc(QARSH_BUFSIZE); memset(b_out, 0, QARSH_BUFSIZE); z_out = 0;
	b_err = malloc(QARSH_BUFSIZE); memset(b_err, 0, QARSH_BUFSIZE); z_err = 0;

	hbeat(qarsh_hb);
	cmd_finished = 0;
	for (;;) {
		if (cmd_finished) {
			timeout.tv_sec = 1;
			timeout.tv_nsec = 0;
		} else {
			timeout.tv_sec = 5;
			timeout.tv_nsec = 0;
		}
		testfds = readfds;
		FD_ZERO(&testwds);
		/* If there are things to write, set the write bit and clear the read */
		/* We use one buffer per channel and we don't want to overwrite it */
		if (z_in) { FD_CLR(fileno(stdin), &testfds); FD_SET(c_in, &testwds); }
		if (z_out) { FD_CLR(c_out, &testfds); FD_SET(fileno(stdout), &testwds); }
		if (z_err) { FD_CLR(c_err, &testfds); FD_SET(fileno(stderr), &testwds); }

		nset = pselect(FD_SETSIZE, &testfds, &testwds, NULL, &timeout, 
				&pselect_sigmask);

		if (nset == 0) {
			if (!hbeat(qarsh_hb)) {
				/* If the heartbeat fails, we should exit now.
				 * The hbeat state will take precedence over the
				 * exit status of the command. */
				break;
			}
		}

		if (nset == -1 && errno == EINTR) {
			/* Only test signals */
			if (signal_to_send) {
				qp = make_qp_kill(signal_to_send);
				send_packet(qarsh_fd, qp);
				qpfree(qp);
				signal_to_send = 0;
			}
		} else if (nset > 0) {
			/* We got traffic, poke the state into the hbeat because
			 * we may not have gotten the chance to call hbeat() above
			 * which would normally reset the state for us.
			 */
			hbeat_setstate(qarsh_hb, HOST_ALIVE);
			
			if (nset && FD_ISSET(fileno(stdin), &testfds)) {
				z_in = read(fileno(stdin), b_in, QARSH_BUFSIZE);
				if (z_in == 0) {
					FD_CLR(fileno(stdin), &readfds);
					close(fileno(stdin));
					close(c_in);
					c_in = 0;
				}
				nset--;
			}
			if (nset && c_in && FD_ISSET(c_in, &testwds)) {
				write(c_in, b_in, z_in);
				z_in = 0;
				memset(b_in, 0, QARSH_BUFSIZE);
			}
			if (nset && c_out && FD_ISSET(c_out, &testfds)) {
				z_out = read(c_out, b_out, QARSH_BUFSIZE);
				if (z_out == 0) {
					FD_CLR(c_out, &readfds);
					close(c_out);
					c_out = 0;
				}
				nset--;
			}
			if (nset && FD_ISSET(fileno(stdout), &testwds)) {
				write(fileno(stdout), b_out, z_out);
				z_out = 0;
			}
			if (nset && c_err && FD_ISSET(c_err, &testfds)) {
				z_err = read(c_err, b_err, QARSH_BUFSIZE);
				if (z_err == 0) {
					FD_CLR(c_err, &readfds);
					close(c_err);
					c_err = 0;
				}
				nset--;
			}
			if (nset && FD_ISSET(fileno(stderr), &testwds)) {
				write(fileno(stderr), b_err, z_err);
				z_err = 0;
			}
			if (nset && FD_ISSET(qarsh_fd, &testfds)) {
				qp = recv_packet(qarsh_fd);
				if (qp == NULL) {
					fprintf(stderr, "recv_packet() returned NULL!\n:");
					break;
				}
			
				/* dump_qp(qp); */
				if (qp && qp->qp_type == QP_CMDEXIT) {
					cmd_finished = 1;
					rc = qp->qp_cmdexit.qp_status;
					qpfree(qp);
					/* Don't break yet, we need to make
					 * sure all output is read. */
				}
				nset--;
			}
		}
		if (cmd_finished && c_out == 0 && c_err == 0) {
			/* If the command is complete and both output
			 * sockets are closed, we can exit now.  We need
			 * to test all conditions at once so if none are
			 * true, we'll still check for heartbeat. */
			break;
		}
	}
	if (c_out) close(c_out);
	if (c_err) close(c_err);
	free(b_in); free(b_out); free(b_err);
	if (hbeat_getstate(qarsh_hb) == HOST_TIMEOUT) {
		fprintf(stderr, "Didn't receive heartbeat for %d seconds\n", 
				hbeat_getmaxtimeout(qarsh_hb));
		return W_EXITCODE(127, 0);
	} else if (hbeat_getstate(qarsh_hb) == HOST_REBOOT) {
		fprintf(stderr, "Remote host rebooted\n");
		return W_EXITCODE(127, 0);
	}
	if (!cmd_finished) {
		fprintf(stderr, "Remote command exited with unknown state\n");
		return W_EXITCODE(127, 0);
	}
	return rc;
}

int 
main(int argc, char *argv[])
{
	int c;
	int port = 5008;
	char *host = NULL;
	char *remuser = NULL;
	char *remgroup = NULL;
	char *args;
	struct passwd *pw;
	int ret;
	struct sigaction sa;
	sigset_t sigmask;
	char *cp;
	int max_timeout = 120;

	openlog("qarsh", LOG_PID, LOG_DAEMON);

	if ((cp = getenv("QARSH_TIMEOUT")) != NULL) {
		max_timeout = atoi(cp);
	}

again:
	while ((c = getopt(argc, argv, "+p:l:g:t:")) != -1) {
	switch (c) {
		case 'l':
			remuser = strdup(optarg);
			break;
		case 'g':
			remgroup = strdup(optarg);
			break;
		case 'p':
			port = atoi(optarg);
			break;
		case 't':
			max_timeout = atoi(optarg);
			break;
		case '?':
		default:
			printf("Unknown option %c\n", (char)optopt);
			usage();
			exit(1);
	} }
	/* Some programs (rsync) put the hostname before some qarsh options
	 * We need to pull the hostname out and if there are still args
	 * keep trying to parse them.  (code from OpenSSH)
	 */
	argc -= optind;
	argv += optind;

	if (argc > 0 && !host && **argv != '-') {
		host = *argv;
		if (argc > 1) {
			optind = 1;
			goto again;
		}
		argc--;
		argv++;
	}

	if (!host) {
		usage();
		exit(1);
	}

	/* check for user and group in form [user[.group]@]hostname */
	{
		char *sp;

		if ((sp = strchr(host, '@'))) {
			remuser = host;
			host = sp+1;
			*sp = '\0';
		}
		if (remuser && (sp = strchr(remuser, '.'))) {
			remgroup = sp+1;
			*sp = '\0';
		}
	}
	if (remuser == NULL) {
		if (!(pw = getpwuid(getuid()))) {
			fprintf(stderr, "qarsh: can not look up local username.\n");
			exit(1);
		}
		remuser = strdup(pw->pw_name);
	}

	if ((args = copyargs(argv)) == NULL) {
		usage();
		exit(1);
	}

	memset(&sa, 0, sizeof sa);
	sigemptyset(&sigmask);
	sa.sa_mask = sigmask;
	sa.sa_flags = SA_ONESHOT;
	sa.sa_handler = sig_alrm_handler;
	sigaction(SIGALRM, &sa, NULL);

	alarm(max_timeout);
	qarsh_fd = connect_to_host(host, port);
	alarm(0);

	if (qarsh_fd == -1) {
		if (errno == 0) {
			fprintf(stderr, "Could not connect to %s:%d, %d: %s\n",
				host, port, h_errno, hstrerror(h_errno));
		} else {
			fprintf(stderr, "Could not connect to %s:%d, %d: %s\n",
				host, port, errno, strerror(errno));
		}
		return 127;
	}

	qarsh_hb = hbeat_init(host, max_timeout);
	if (!qarsh_hb) {
		fprintf(stderr, "Could not init heartbeat to %s\n", host);
		return 127;
	}

	set_remote_user(remuser, remgroup);

	ret = run_remote_cmd(args);
	close(qarsh_fd);
	free(args);
	/* If the remote cmd was killed, we need to be killed too */
	if (WIFSIGNALED(ret)) {
		reset_signals();
		raise(WTERMSIG(ret));
	}
	/* Otherwise we need to exit with the same exit status */
	return WEXITSTATUS(ret);
}
